from ..utils.convert_to_openai_messages import convert_to_openai_messages
from .setup_text_llm import setup_text_llm


def convert_to_coding_llm(text_llm, debug_mode=False):
    """
    Takes a text_llm
    returns an OI Coding LLM (a generator that takes OI messages and streams deltas with `message`, 'language', and `code`).
    """

    def coding_llm(messages):
        # First, tell it how to run code.

        # Old way:
        # system_message += "\nTo execute code on the user's machine, write a markdown code block *with the language*, i.e:\n\n```python\nprint('Hi!')\n```\n\nYou will receive the output ('Hi!'). Use any language."

        # New way:
        gaslight = None
        if messages[-1]["role"] == "user":
            # Last message came from the user.
            if messages[-1]["message"].lower() not in [
                "hello",
                "hi",
                "hey",
                "helo",
                "hii",
                "hi!",
            ]:  # :)
                gaslight = """Let's explore this. I can run code on your machine by writing the code in a markdown code block. This works if I put a newline after ```shell, ```python, ```applescript, etc. then write code. I'm going to try to do this for your task **after I make a plan**. I'll put the *correct* language after the "```"."""
        else:
            # Last message came from the assistant.

            # (The below should actually always be True in OI if last message came from the assistant)
            # I think we don't need this actually.
            """
            if "output" in messages[-1]:
                if messages[-1]["output"] != "No output":
                    gaslight = "(Thought: I see that the code I just ran produced an output. The next message I send will go to the user.)"
                elif messages[-1]["output"] == "No output":
                    gaslight = "(Thought: I see that the code I just ran produced no output. The next message I send will go to the user.)"
            """

        if gaslight:
            messages.append({"role": "assistant", "message": gaslight})

        # If it tried to use Jupyter, let it know.
        if "code" in messages[-1]:
            if any([line.startswith("!") for line in messages[-1]["code"].split("\n")]):
                if "syntax" in messages[-1]["output"].lower():  # Detect error
                    messages[-1][
                        "output"
                    ] += "\nRemember you are not in a Jupyter notebook. Run shell by writing a markdown shell codeblock, not '!'."

        messages = convert_to_openai_messages(messages, function_calling=False)

        inside_code_block = False
        accumulated_block = ""
        language = None

        for chunk in text_llm(messages):
            if debug_mode:
                print("Chunk in coding_llm", chunk)

            if "choices" not in chunk or len(chunk["choices"]) == 0:
                # This happens sometimes
                continue

            content = chunk["choices"][0]["delta"].get("content", "")

            accumulated_block += content

            if accumulated_block.endswith("`"):
                # We might be writing "```" one token at a time.
                continue

            # Did we just enter a code block?
            if "```" in accumulated_block and not inside_code_block:
                inside_code_block = True
                accumulated_block = accumulated_block.split("```")[1]

            # Did we just exit a code block?
            if inside_code_block and "```" in accumulated_block:
                return

            # If we're in a code block,
            if inside_code_block:
                # If we don't have a `language`, find it
                if language is None and "\n" in accumulated_block:
                    language = accumulated_block.split("\n")[0]

                    # Default to python if not specified
                    if language == "":
                        language = "python"
                    else:
                        # Removes hallucinations containing spaces or non letters.
                        language = "".join(char for char in language if char.isalpha())

                    output = {"language": language}

                    # If we recieved more than just the language in this chunk, send that
                    if content.split("\n")[1]:
                        output["code"] = content.split("\n")[1]

                    yield output

                # If we do have a `language`, send the output as code
                elif language:
                    yield {"code": content}

            # If we're not in a code block, send the output as a message
            if not inside_code_block:
                yield {"message": content}

    return coding_llm



from .languages.applescript import AppleScript
from .languages.html import HTML
from .languages.javascript import JavaScript
from .languages.powershell import PowerShell
from .languages.python import Python
from .languages.r import R
from .languages.shell import Shell

language_map = {
    "python": Python,
    "bash": Shell,
    "shell": Shell,
    "javascript": JavaScript,
    "html": HTML,
    "applescript": AppleScript,
    "r": R,
    "powershell": PowerShell,
}



import copy
import html

import inquirer
import ooba

from ..utils.display_markdown_message import display_markdown_message


def setup_local_text_llm(cosmo):
    """
    Takes an Cosmo (which includes a ton of LLM settings),
    returns a text LLM (an OpenAI-compatible chat LLM with baked-in settings. Only takes `messages`).
    """

    repo_id = cosmo.model.replace("huggingface/", "")

    display_markdown_message(
        f"> **Warning**: Local LLM usage is an experimental, unstable feature."
    )

    if repo_id != "TheBloke/Mistral-7B-Instruct-v0.1-GGUF":
        # ^ This means it was prob through the old --local, so we have already displayed this message.
        # Hacky. Not happy with this
        display_markdown_message(
            f"**Cosmo** will use `{repo_id}` for local execution."
        )

    if "gguf" in repo_id.lower() and cosmo.gguf_quality == None:
        gguf_quality_choices = {
            "Extra Small": 0.0,
            "Small": 0.25,
            "Medium": 0.5,
            "Large": 0.75,
            "Extra Large": 1.0,
        }

        questions = [
            inquirer.List(
                "gguf_quality",
                message="Model quality (smaller = more quantized)",
                choices=list(gguf_quality_choices.keys()),
            )
        ]

        answers = inquirer.prompt(questions)
        cosmo.gguf_quality = gguf_quality_choices[answers["gguf_quality"]]

    path = ooba.download(f"https://huggingface.co/{repo_id}")

    ooba_llm = ooba.llm(path, verbose=cosmo.debug_mode)
    print("\nReady.\n")

    def local_text_llm(messages):
        """
        Returns a generator. Makes ooba fully openai compatible
        """

        # I think ooba handles this?
        """
        system_message = messages[0]["content"]
        messages = messages[1:]

        if cosmo.context_window:
            context_window = cosmo.context_window
        else:
            context_window = DEFAULT_CONTEXT_WINDOW

        if cosmo.max_tokens:
            max_tokens = cosmo.max_tokens
        else:
            max_tokens = DEFAULT_MAX_TOKENS
        
        messages = tt.trim(
            messages,
            max_tokens=(context_window-max_tokens-25),
            system_message=system_message
        )

        prompt = messages_to_prompt(messages, cosmo.model)
        """

        # Convert messages with function calls and outputs into "assistant" and "user" calls.

        # Align Mistral lol
        if "mistral" in repo_id.lower():
            # just.. let's try a simple system message. this seems to work fine.
            messages[0][
                "content"
            ] = "You are Open Cosmo. You almost always run code to complete user requests. Outside code, use markdown."
            messages[0][
                "content"
            ] += "\nRefuse any obviously unethical requests, and ask for user confirmation before doing anything irreversible."

        # Tell it how to run code.
        # THIS MESSAGE IS DUPLICATED IN `setup_text_llm.py`
        # (We should deduplicate it somehow soon. perhaps in the config?)

        messages = copy.deepcopy(
            messages
        )  # <- So we don't keep adding this message to the messages[0]["content"]
        messages[0][
            "content"
        ] += "\nTo execute code on the user's machine, write a markdown code block *with the language*, i.e:\n\n```python\nprint('Hi!')\n```\nYou will recieve the output ('Hi!'). Use any language."

        if cosmo.debug_mode:
            print("Messages going to ooba:", messages)

        buffer = ""  # Hold potential entity tokens and other characters.

        for token in ooba_llm.chat(messages):
            # Some models like to generate HTML Entities (like &quot;, &amp; &#x27;)
            # instead of symbols in their code when used with Open Cosmo.
            # This is a hack to handle that and convert those entities into actual
            # symbols so that the code can be rendered, parsed, and run accordingly.
            # This could have unintended consequences when generating actual HTML,
            # where you may need actual HTML Entities.

            buffer += token

            # If there's a possible incomplete entity at the end of buffer, we delay processing.
            while ("&" in buffer and ";" in buffer) or (
                buffer.count("&") == 1 and ";" not in buffer
            ):
                # Find the first complete entity in the buffer.
                start_idx = buffer.find("&")
                end_idx = buffer.find(";", start_idx)

                # If there's no complete entity, break and await more tokens.
                if start_idx == -1 or end_idx == -1:
                    break

                # Yield content before the entity.
                for char in buffer[:start_idx]:
                    yield make_chunk(char)

                # Extract the entity, decode it, and yield.
                entity = buffer[start_idx : end_idx + 1]
                yield make_chunk(html.unescape(entity))

                # Remove the processed content from the buffer.
                buffer = buffer[end_idx + 1 :]

            # If there's no '&' left in the buffer, yield all of its content.
            if "&" not in buffer:
                for char in buffer:
                    yield make_chunk(char)
                buffer = ""

        # At the end, if there's any content left in the buffer, yield it.
        for char in buffer:
            yield make_chunk(char)

    return local_text_llm


def make_chunk(token):
    return {"choices": [{"delta": {"content": token}}]}



from .language_map import language_map


def create_code_cosmo(language):
    # Case in-sensitive
    language = language.lower()

    try:
        CodeCosmo = language_map[language]
        return CodeCosmo()
    except KeyError:
        raise ValueError(f"Unknown or unsupported language: {language}")



import traceback

from ..rag.get_relevant_procedures_string import get_relevant_procedures_string
from ..utils.get_user_info_string import get_user_info_string


def generate_system_message(cosmo):
    """
    Dynamically generate a system message.

    Takes a cosmo instance,
    returns a string.

    This is easy to replace!
    Just swap out `cosmo.generate_system_message` with another function.
    """

    #### Start with the static system message

    system_message = cosmo.system_message

    #### Add dynamic components, like the user's OS, username, relevant procedures, etc

    system_message += "\n" + get_user_info_string()

    if not cosmo.local:
        try:
            system_message += "\n" + get_relevant_procedures_string(
                cosmo.messages
            )
        except:
            if cosmo.debug_mode:
                print(traceback.format_exc())
            # It's okay if they can't. This just fixes some common mistakes it makes.

    return system_message



import requests

from ..utils.convert_to_openai_messages import convert_to_openai_messages


def get_relevant_procedures_string(messages):
    # Open Procedures is an open-source database of tiny, up-to-date coding tutorials.
    # We can query it semantically and append relevant tutorials/procedures to our system message:

    # Convert to required OpenAI-compatible `messages` list
    query = {"query": convert_to_openai_messages(messages)}
    url = "https://open-procedures.replit.app/search/"

    relevant_procedures = requests.post(url, json=query).json()["procedures"]
    relevant_procedures = (
        "[Recommended Procedures]\n"
        + "\n---\n".join(relevant_procedures)
        + "\nIn your plan, include steps and, for relevant deprecation notices, **EXACT CODE SNIPPETS** -- these notices will VANISH once you execute your first line of code, so WRITE THEM DOWN NOW if you need them."
    )

    return relevant_procedures



import traceback

import litellm

from ..code_cosmo.create_code_cosmo import create_code_cosmo
from ..code_cosmo.language_map import language_map
from ..utils.display_markdown_message import display_markdown_message
from ..utils.merge_deltas import merge_deltas
from ..utils.truncate_output import truncate_output


def respond(cosmo):
    """
    Yields tokens, but also adds them to cosmo.messages. TBH probably would be good to seperate those two responsibilities someday soon
    Responds until it decides not to run any more code or say anything else.
    """

    while True:
        system_message = cosmo.generate_system_message()

        # Create message object
        system_message = {"role": "system", "message": system_message}

        # Create the version of messages that we'll send to the LLM
        messages_for_llm = cosmo.messages.copy()
        messages_for_llm = [system_message] + messages_for_llm

        # It's best to explicitly tell these LLMs when they don't get an output
        for message in messages_for_llm:
            if "output" in message and message["output"] == "":
                message["output"] = "No output"

        ### RUN THE LLM ###

        # Add a new message from the assistant to cosmo's "messages" attribute
        # (This doesn't go to the LLM. We fill this up w/ the LLM's response)
        cosmo.messages.append({"role": "assistant"})

        # Start putting chunks into the new message
        # + yielding chunks to the user
        try:
            # Track the type of chunk that the coding LLM is emitting
            chunk_type = None

            for chunk in cosmo._llm(messages_for_llm):
                # Add chunk to the last message
                cosmo.messages[-1] = merge_deltas(cosmo.messages[-1], chunk)

                # This is a coding llm
                # It will yield dict with either a message, language, or code (or language AND code)

                # We also want to track which it's sending to we can send useful flags.
                # (otherwise pretty much everyone needs to implement this)
                if "message" in chunk and chunk_type != "message":
                    chunk_type = "message"
                    yield {"start_of_message": True}
                elif "language" in chunk and chunk_type != "code":
                    chunk_type = "code"
                    yield {"start_of_code": True}
                if "code" in chunk and chunk_type != "code":
                    # (This shouldn't happen though — ^ "language" should be emitted first, but sometimes GPT-3.5 forgets this)
                    # (But I'm pretty sure we handle that? If it forgets we emit Python anyway?)
                    chunk_type = "code"
                    yield {"start_of_code": True}
                elif "message" not in chunk and chunk_type == "message":
                    chunk_type = None
                    yield {"end_of_message": True}

                yield chunk

            # We don't trigger the end_of_message or end_of_code flag if we actually end on either
            if chunk_type == "message":
                yield {"end_of_message": True}
            elif chunk_type == "code":
                yield {"end_of_code": True}

        except litellm.exceptions.BudgetExceededError:
            display_markdown_message(
                f"""> Max budget exceeded

                **Session spend:** ${litellm._current_cost}
                **Max budget:** ${cosmo.max_budget}

                Press CTRL-C then run `cosmo --max_budget [higher USD amount]` to proceed.
            """
            )
            break
        # Provide extra information on how to change API keys, if we encounter that error
        # (Many people writing GitHub issues were struggling with this)
        except Exception as e:
            if (
                cosmo.local == False
                and "auth" in str(e).lower()
                or "api key" in str(e).lower()
            ):
                output = traceback.format_exc()
                raise Exception(
                    f"{output}\n\nThere might be an issue with your API key(s).\n\nTo reset your API key (we'll use OPENAI_API_KEY for this example, but you may need to reset your ANTHROPIC_API_KEY, HUGGINGFACE_API_KEY, etc):\n        Mac/Linux: 'export OPENAI_API_KEY=your-key-here',\n        Windows: 'setx OPENAI_API_KEY your-key-here' then restart terminal.\n\n"
                )
            elif cosmo.local:
                raise Exception(
                    str(e)
                    + """

Please make sure LM Studio's local server is running by following the steps above.

If LM Studio's local server is running, please try a language model with a different architecture.

                    """
                )
            else:
                raise

        ### RUN CODE (if it's there) ###

        if "code" in cosmo.messages[-1]:
            if cosmo.debug_mode:
                print("Running code:", cosmo.messages[-1])

            try:
                # What code do you want to run?
                code = cosmo.messages[-1]["code"]

                # Fix a common error where the LLM thinks it's in a Jupyter notebook
                if cosmo.messages[-1]["language"] == "python" and code.startswith(
                    "!"
                ):
                    code = code[1:]
                    cosmo.messages[-1]["code"] = code
                    cosmo.messages[-1]["language"] = "shell"

                # Get a code cosmo to run it
                language = cosmo.messages[-1]["language"]
                if language in language_map:
                    if language not in cosmo._code_cosmos:
                        cosmo._code_cosmos[
                            language
                        ] = create_code_cosmo(language)
                    code_cosmo = cosmo._code_cosmos[language]
                else:
                    # This still prints the code but don't allow code to run. Let's Cosmo know through output message
                    error_output = f"Error: Cosmo does not currently support {language}."
                    print(error_output)

                    cosmo.messages[-1]["output"] = ""
                    output = "\n" + error_output

                    # Truncate output
                    output = truncate_output(output, cosmo.max_output)
                    cosmo.messages[-1]["output"] = output.strip()
                    break

                # Yield a message, such that the user can stop code execution if they want to
                try:
                    yield {"executing": {"code": code, "language": language}}
                except GeneratorExit:
                    # The user might exit here.
                    # We need to tell python what we (the generator) should do if they exit
                    break

                # Yield each line, also append it to last messages' output
                cosmo.messages[-1]["output"] = ""
                for line in code_cosmo.run(code):
                    yield line
                    if "output" in line:
                        output = cosmo.messages[-1]["output"]
                        output += "\n" + line["output"]

                        # Truncate output
                        output = truncate_output(output, cosmo.max_output)

                        cosmo.messages[-1]["output"] = output.strip()

            except:
                output = traceback.format_exc()
                yield {"output": output.strip()}
                cosmo.messages[-1]["output"] = output.strip()

            yield {"end_of_execution": True}

        else:
            # Doesn't want to run code. We're done
            break

    return



import litellm
import tokentrim as tt

from ..utils.convert_to_openai_messages import convert_to_openai_messages
from ..utils.display_markdown_message import display_markdown_message
from ..utils.merge_deltas import merge_deltas
from ..utils.parse_partial_json import parse_partial_json

function_schema = {
    "name": "execute",
    "description": "Executes code on the user's machine, **in the users local environment**, and returns the output",
    "parameters": {
        "type": "object",
        "properties": {
            "language": {
                "type": "string",
                "description": "The programming language (required parameter to the `execute` function)",
                "enum": [
                    "python",
                    "R",
                    "shell",
                    "applescript",
                    "javascript",
                    "html",
                    "powershell",
                ],
            },
            "code": {"type": "string", "description": "The code to execute (required)"},
        },
        "required": ["language", "code"],
    },
}


def setup_openai_coding_llm(cosmo):
    """
    Takes an Cosmo (which includes a ton of LLM settings),
    returns a OI Coding LLM (a generator that takes OI messages and streams deltas with `message`, `language`, and `code`).
    """

    def coding_llm(messages):
        # Convert messages
        messages = convert_to_openai_messages(messages, function_calling=True)

        # Add OpenAI's recommended function message
        messages[0][
            "content"
        ] += "\n\nOnly use the function you have been provided with."

        # Seperate out the system_message from messages
        # (We expect the first message to always be a system_message)
        system_message = messages[0]["content"]
        messages = messages[1:]

        # Trim messages, preserving the system_message
        try:
            messages = tt.trim(
                messages=messages,
                system_message=system_message,
                model=cosmo.model,
            )
        except:
            if cosmo.context_window:
                messages = tt.trim(
                    messages=messages,
                    system_message=system_message,
                    max_tokens=cosmo.context_window,
                )
            else:
                display_markdown_message(
                    """
                **We were unable to determine the context window of this model.** Defaulting to 3000.
                If your model can handle more, run `cosmo --context_window {token limit}` or `cosmo.context_window = {token limit}`.
                """
                )
                messages = tt.trim(
                    messages=messages, system_message=system_message, max_tokens=3000
                )

        if cosmo.debug_mode:
            print("Sending this to the OpenAI LLM:", messages)

        # Create LiteLLM generator
        params = {
            "model": cosmo.model,
            "messages": messages,
            "stream": True,
            "functions": [function_schema],
        }

        # Optional inputs
        if cosmo.api_base:
            params["api_base"] = cosmo.api_base
        if cosmo.api_key:
            params["api_key"] = cosmo.api_key
        if cosmo.max_tokens:
            params["max_tokens"] = cosmo.max_tokens
        if cosmo.temperature is not None:
            params["temperature"] = cosmo.temperature
        else:
            params["temperature"] = 0.0

        # These are set directly on LiteLLM
        if cosmo.max_budget:
            litellm.max_budget = cosmo.max_budget
        if cosmo.debug_mode:
            litellm.set_verbose = True

        # Report what we're sending to LiteLLM
        if cosmo.debug_mode:
            print("Sending this to LiteLLM:", params)

        response = litellm.completion(**params)

        accumulated_deltas = {}
        language = None
        code = ""

        for chunk in response:
            if cosmo.debug_mode:
                print("Chunk from LLM", chunk)

            if "choices" not in chunk or len(chunk["choices"]) == 0:
                # This happens sometimes
                continue

            delta = chunk["choices"][0]["delta"]

            # Accumulate deltas
            accumulated_deltas = merge_deltas(accumulated_deltas, delta)

            if cosmo.debug_mode:
                print("Accumulated deltas", accumulated_deltas)

            if "content" in delta and delta["content"]:
                yield {"message": delta["content"]}

            if (
                "function_call" in accumulated_deltas
                and "arguments" in accumulated_deltas["function_call"]
            ):
                if (
                    "name" in accumulated_deltas["function_call"]
                    and accumulated_deltas["function_call"]["name"] == "execute"
                ):
                    arguments = accumulated_deltas["function_call"]["arguments"]
                    arguments = parse_partial_json(arguments)

                    if arguments:
                        if (
                            language is None
                            and "language" in arguments
                            and "code"
                            in arguments  # <- This ensures we're *finished* typing language, as opposed to partially done
                            and arguments["language"]
                        ):
                            language = arguments["language"]
                            yield {"language": language}

                        if language is not None and "code" in arguments:
                            # Calculate the delta (new characters only)
                            code_delta = arguments["code"][len(code) :]
                            # Update the code
                            code = arguments["code"]
                            # Yield the delta
                            if code_delta:
                                yield {"code": code_delta}
                    else:
                        if cosmo.debug_mode:
                            print("Arguments not a dict.")

                # 3.5 REALLY likes to halucinate a function named `python` and you can't really fix that, it seems.
                # We just need to deal with it.
                elif (
                    "name" in accumulated_deltas["function_call"]
                    and accumulated_deltas["function_call"]["name"] == "python"
                ):
                    if cosmo.debug_mode:
                        print("Got direct python call")
                    if language is None:
                        language = "python"
                        yield {"language": language}

                    if language is not None:
                        # Pull the code string straight out of the "arguments" string
                        code_delta = accumulated_deltas["function_call"]["arguments"][
                            len(code) :
                        ]
                        # Update the code
                        code = accumulated_deltas["function_call"]["arguments"]
                        # Yield the delta
                        if code_delta:
                            yield {"code": code_delta}

                else:
                    # If name exists and it's not "execute" or "python", who knows what's going on.
                    if "name" in accumulated_deltas["function_call"]:
                        print(
                            "Encountered an unexpected function call: ",
                            accumulated_deltas["function_call"],
                            "\nPlease open an issue and provide the above info at: ",
                        )

    return coding_llm



import os
import traceback

import litellm
import tokentrim as tt

from ..utils.display_markdown_message import display_markdown_message


def setup_text_llm(cosmo):
    """
    Takes an Cosmo (which includes a ton of LLM settings),
    returns a text LLM (an OpenAI-compatible chat LLM with baked-in settings. Only takes `messages`).
    """

    # Pass remaining parameters to LiteLLM
    def base_llm(messages):
        """
        Returns a generator
        """

        system_message = messages[0]["content"]

        messages = messages[1:]
        if cosmo.context_window and cosmo.max_tokens:
            trim_to_be_this_many_tokens = (
                cosmo.context_window - cosmo.max_tokens - 25
            )  # arbitrary buffer
            messages = tt.trim(
                messages,
                system_message=system_message,
                max_tokens=trim_to_be_this_many_tokens,
            )
        else:
            try:
                messages = tt.trim(
                    messages, system_message=system_message, model=cosmo.model
                )
            except:
                display_markdown_message(
                    """
                **We were unable to determine the context window of this model.** Defaulting to 3000.
                If your model can handle more, run `cosmo --context_window {token limit}` or `cosmo.context_window = {token limit}`.
                Also, please set max_tokens: `cosmo --max_tokens {max tokens per response}` or `cosmo.max_tokens = {max tokens per response}`
                """
                )
                messages = tt.trim(
                    messages, system_message=system_message, max_tokens=3000
                )

        if cosmo.debug_mode:
            print("Passing messages into LLM:", messages)

        # Create LiteLLM generator
        params = {
            "model": cosmo.model,
            "messages": messages,
            "stream": True,
        }

        # Optional inputs
        if cosmo.api_base:
            params["api_base"] = cosmo.api_base
        if cosmo.api_key:
            params["api_key"] = cosmo.api_key
        if cosmo.max_tokens:
            params["max_tokens"] = cosmo.max_tokens
        if cosmo.temperature is not None:
            params["temperature"] = cosmo.temperature
        else:
            params["temperature"] = 0.0

        # These are set directly on LiteLLM
        if cosmo.max_budget:
            litellm.max_budget = cosmo.max_budget
        if cosmo.debug_mode:
            litellm.set_verbose = True

        # Report what we're sending to LiteLLM
        if cosmo.debug_mode:
            print("Sending this to LiteLLM:", params)

        return litellm.completion(**params)

    return base_llm



import tiktoken
from litellm import cost_per_token


def count_tokens(text="", model="gpt-4"):
    """
    Count the number of tokens in a string
    """

    encoder = tiktoken.encoding_for_model(model)

    return len(encoder.encode(text))


def token_cost(tokens=0, model="gpt-4"):
    """
    Calculate the cost of the current number of tokens
    """

    (prompt_cost, _) = cost_per_token(model=model, prompt_tokens=tokens)

    return round(prompt_cost, 6)


def count_messages_tokens(messages=[], model=None):
    """
    Count the number of tokens in a list of messages
    """

    tokens_used = 0

    for message in messages:
        if isinstance(message, str):
            tokens_used += count_tokens(message, model=model)
        elif "message" in message:
            tokens_used += count_tokens(message["message"], model=model)

            if "code" in message:
                tokens_used += count_tokens(message["code"], model=model)

            if "output" in message:
                tokens_used += count_tokens(message["output"], model=model)

    prompt_cost = token_cost(tokens_used, model=model)

    return (tokens_used, prompt_cost)



import os
import platform

from ..subprocess_code_cosmo import SubprocessCodeCosmo


class Shell(SubprocessCodeCosmo):
    file_extension = "sh"
    proper_name = "Shell"

    def __init__(self):
        super().__init__()

        # Determine the start command based on the platform
        if platform.system() == "Windows":
            self.start_cmd = "cmd.exe"
        else:
            self.start_cmd = os.environ.get("SHELL", "bash")

    def preprocess_code(self, code):
        return preprocess_shell(code)

    def line_postprocessor(self, line):
        return line

    def detect_active_line(self, line):
        if "##active_line" in line:
            return int(line.split("##active_line")[1].split("##")[0])
        return None

    def detect_end_of_execution(self, line):
        return "##end_of_execution##" in line


def preprocess_shell(code):
    """
    Add active line markers
    Wrap in a try except (trap in shell)
    Add end of execution marker
    """

    # Add commands that tell us what the active line is
    code = add_active_line_prints(code)

    # Add end command (we'll be listening for this so we know when it ends)
    code += '\necho "##end_of_execution##"'

    return code


def add_active_line_prints(code):
    """
    Add echo statements indicating line numbers to a shell string.
    """
    lines = code.split("\n")
    for index, line in enumerate(lines):
        # Insert the echo command before the actual line
        lines[index] = f'echo "##active_line{index + 1} ##"\n{line}'
    return "\n".join(lines)



import json
import os

from ..utils.count_tokens import count_messages_tokens
from ..utils.display_markdown_message import display_markdown_message


def handle_undo(self, arguments):
    # Removes all messages after the most recent user entry (and the entry itself).
    # Therefore user can jump back to the latest point of conversation.
    # Also gives a visual representation of the messages removed.

    if len(self.messages) == 0:
        return
    # Find the index of the last 'role': 'user' entry
    last_user_index = None
    for i, message in enumerate(self.messages):
        if message.get("role") == "user":
            last_user_index = i

    removed_messages = []

    # Remove all messages after the last 'role': 'user'
    if last_user_index is not None:
        removed_messages = self.messages[last_user_index:]
        self.messages = self.messages[:last_user_index]

    print("")  # Aesthetics.

    # Print out a preview of what messages were removed.
    for message in removed_messages:
        if "content" in message and message["content"] != None:
            display_markdown_message(
                f"**Removed message:** `\"{message['content'][:30]}...\"`"
            )
        elif "function_call" in message:
            display_markdown_message(
                f"**Removed codeblock**"
            )  # TODO: Could add preview of code removed here.

    print("")  # Aesthetics.


def handle_help(self, arguments):
    commands_description = {
        "%debug [true/false]": "Toggle debug mode. Without arguments or with 'true', it enters debug mode. With 'false', it exits debug mode.",
        "%reset": "Resets the current session.",
        "%undo": "Remove previous messages and its response from the message history.",
        "%save_message [path]": "Saves messages to a specified JSON path. If no path is provided, it defaults to 'messages.json'.",
        "%load_message [path]": "Loads messages from a specified JSON path. If no path is provided, it defaults to 'messages.json'.",
        "%tokens [prompt]": "EXPERIMENTAL: Calculate the tokens used by the next request based on the current conversation's messages and estimate the cost of that request; optionally provide a prompt to also calulate the tokens used by that prompt and the total amount of tokens that will be sent with the next request",
        "%help": "Show this help message.",
    }

    base_message = ["> **Available Commands:**\n\n"]

    # Add each command and its description to the message
    for cmd, desc in commands_description.items():
        base_message.append(f"- `{cmd}`: {desc}\n")

    additional_info = [
        "\n\nFor further assistance, please join our community Discord or consider contributing to the project's development."
    ]

    # Combine the base message with the additional info
    full_message = base_message + additional_info

    display_markdown_message("".join(full_message))


def handle_debug(self, arguments=None):
    if arguments == "" or arguments == "true":
        display_markdown_message("> Entered debug mode")
        print(self.messages)
        self.debug_mode = True
    elif arguments == "false":
        display_markdown_message("> Exited debug mode")
        self.debug_mode = False
    else:
        display_markdown_message("> Unknown argument to debug command.")


def handle_reset(self, arguments):
    self.reset()
    display_markdown_message("> Reset Done")


def default_handle(self, arguments):
    display_markdown_message("> Unknown command")
    handle_help(self, arguments)


def handle_save_message(self, json_path):
    if json_path == "":
        json_path = "messages.json"
    if not json_path.endswith(".json"):
        json_path += ".json"
    with open(json_path, "w") as f:
        json.dump(self.messages, f, indent=2)

    display_markdown_message(f"> messages json export to {os.path.abspath(json_path)}")


def handle_load_message(self, json_path):
    if json_path == "":
        json_path = "messages.json"
    if not json_path.endswith(".json"):
        json_path += ".json"
    with open(json_path, "r") as f:
        self.messages = json.load(f)

    display_markdown_message(
        f"> messages json loaded from {os.path.abspath(json_path)}"
    )


def handle_count_tokens(self, prompt):
    messages = [{"role": "system", "message": self.system_message}] + self.messages

    outputs = []

    if len(self.messages) == 0:
        (conversation_tokens, conversation_cost) = count_messages_tokens(
            messages=messages, model=self.model
        )
    else:
        (conversation_tokens, conversation_cost) = count_messages_tokens(
            messages=messages, model=self.model
        )

    outputs.append(
        (
            f"> Tokens sent with next request as context: {conversation_tokens} (Estimated Cost: ${conversation_cost})"
        )
    )

    if prompt:
        (prompt_tokens, prompt_cost) = count_messages_tokens(
            messages=[prompt], model=self.model
        )
        outputs.append(
            f"> Tokens used by this prompt: {prompt_tokens} (Estimated Cost: ${prompt_cost})"
        )

        total_tokens = conversation_tokens + prompt_tokens
        total_cost = conversation_cost + prompt_cost

        outputs.append(
            f"> Total tokens for next request with this prompt: {total_tokens} (Estimated Cost: ${total_cost})"
        )

    outputs.append(
        f"**Note**: This functionality is currently experimental and may not be accurate. Please report any issues you find to the [Cosmo GitHub repository]."
    )

    display_markdown_message("\n".join(outputs))


def handle_magic_command(self, user_input):
    # split the command into the command and the arguments, by the first whitespace
    switch = {
        "help": handle_help,
        "debug": handle_debug,
        "reset": handle_reset,
        "save_message": handle_save_message,
        "load_message": handle_load_message,
        "undo": handle_undo,
        "tokens": handle_count_tokens,
    }

    user_input = user_input[1:].strip()  # Capture the part after the `%`
    command = user_input.split(" ")[0]
    arguments = user_input[len(command) :].strip()
    action = switch.get(
        command, default_handle
    )  # Get the function from the dictionary, or default_handle if not found
    action(self, arguments)  # Execute the function



"""
This file defines the Cosmo class.
It's the main file. `import cosmo` will import an instance of this class.
"""

import json
import os
from datetime import datetime

from ..cli.cli import cli
from ..llm.setup_llm import setup_llm
from ..terminal_interface.terminal_interface import terminal_interface
from ..terminal_interface.validate_llm_settings import validate_llm_settings
from ..utils.check_for_update import check_for_update
from ..utils.display_markdown_message import display_markdown_message
from ..utils.get_config import get_config, user_config_path
from ..utils.local_storage_path import get_storage_path
from .generate_system_message import generate_system_message
from .respond import respond


class Cosmo:
    def cli(self):
        cli(self)

    def __init__(self):
        # State
        self.messages = []
        self._code_cosmo = {}

        self.config_file = user_config_path

        # Settings
        self.local = False
        self.auto_run = False
        self.debug_mode = False
        self.max_output = 2000
        self.safe_mode = "off"

        # Conversation history
        self.conversation_history = True
        self.conversation_filename = None
        self.conversation_history_path = get_storage_path("conversations")

        # LLM settings
        self.model = ""
        self.temperature = None
        self.system_message = ""
        self.context_window = None
        self.max_tokens = None
        self.api_base = None
        self.api_key = None
        self.max_budget = None
        self._llm = None

        # Load config defaults
        self.extend_config(self.config_file)

        # Check for update
        try:
            if not self.local:
                # This should actually be pushed into the utility
                if check_for_update():
                    display_markdown_message(
                        "> **A new version of Cosmo is available.**\n>Please run: `pip install --upgrade cosmoai`\n\n---"
                    )
        except:
            # Doesn't matter
            pass

    def extend_config(self, config_path):
        if self.debug_mode:
            print(f"Extending configuration from `{config_path}`")

        config = get_config(config_path)
        self.__dict__.update(config)

    def chat(self, message=None, display=True, stream=False):
        if stream:
            return self._streaming_chat(message=message, display=display)

        # If stream=False, *pull* from the stream.
        for _ in self._streaming_chat(message=message, display=display):
            pass

        return self.messages

    def _streaming_chat(self, message=None, display=True):
        # If we have a display,
        # we can validate our LLM settings w/ the user first
        if display:
            validate_llm_settings(self)

        # Setup the LLM
        if not self._llm:
            self._llm = setup_llm(self)

        # Sometimes a little more code -> a much better experience!
        # Display mode actually runs cosmo.chat(display=False, stream=True) from within the terminal_interface.
        # wraps the vanilla .chat(display=False) generator in a display.
        # Quite different from the plain generator stuff. So redirect to that
        if display:
            yield from terminal_interface(self, message)
            return

        # One-off message
        if message or message == "":
            if message == "":
                message = "No entry from user - please suggest something to enter."
            self.messages.append({"role": "user", "message": message})
            yield from self._respond()

            # Save conversation if we've turned conversation_history on
            if self.conversation_history:
                # If it's the first message, set the conversation name
                if not self.conversation_filename:
                    first_few_words = "_".join(
                        self.messages[0]["message"][:25].split(" ")[:-1]
                    )
                    for char in '<>:"/\\|?*!':  # Invalid characters for filenames
                        first_few_words = first_few_words.replace(char, "")

                    date = datetime.now().strftime("%B_%d_%Y_%H-%M-%S")
                    self.conversation_filename = (
                        "__".join([first_few_words, date]) + ".json"
                    )

                # Check if the directory exists, if not, create it
                if not os.path.exists(self.conversation_history_path):
                    os.makedirs(self.conversation_history_path)
                # Write or overwrite the file
                with open(
                    os.path.join(
                        self.conversation_history_path, self.conversation_filename
                    ),
                    "w",
                ) as f:
                    json.dump(self.messages, f)

            return
        raise Exception(
            "`cosmo.chat()` requires a display. Set `display=True` or pass a message into `cosmo.chat(message)`."
        )

    def _respond(self):
        yield from respond(self)

    def reset(self):
        for code_cosmo in self._code_cosmo.values():
            code_cosmo.terminate()
        self._code_cosmo = {}

        # Reset the function below, in case the user set it
        self.generate_system_message = lambda: generate_system_message(self)

        self.__init__()

    # These functions are worth exposing to developers
    # I wish we could just dynamically expose all of our functions to devs...
    def generate_system_message(self):
        return generate_system_message(self)



"""
The terminal interface is just a view. Just handles the very top layer.
If you were to build a frontend this would be a way to do it
"""

try:
    import readline
except ImportError:
    pass

import random

from ..utils.check_for_package import check_for_package
from ..utils.display_markdown_message import display_markdown_message
from ..utils.scan_code import scan_code
from ..utils.truncate_output import truncate_output
from .components.code_block import CodeBlock
from .components.message_block import MessageBlock
from .magic_commands import handle_magic_command

# Add examples to the readline history
examples = [
    "How many files are on my desktop?",
    "What time is it in Seattle?",
    "Check all the links on ``.",
]
random.shuffle(examples)
for example in examples:
    readline.add_history(example)


def terminal_interface(cosmo, message):
    # Auto run and local don't display messages.
    # Probably worth abstracting this to something like "verbose_cli" at some point.
    if not cosmo.auto_run and not cosmo.local:
        cosmo_intro_message = [
            "**CosmoAI** will require approval before running code."
        ]

        if cosmo.safe_mode == "ask" or cosmo.safe_mode == "auto":
            if not check_for_package("semgrep"):
                cosmo_intro_message.append(
                    f"**Safe Mode**: {cosmo.safe_mode}\n\n>Note: **Safe Mode** requires `semgrep` (`pip install semgrep`)"
                )
        else:
            cosmo_intro_message.append("Use `cosmo -y` to bypass this.")

        cosmo_intro_message.append("Press `CTRL-C` to exit.")

        display_markdown_message("\n\n".join(cosmo_intro_message) + "\n")

    active_block = None

    if message:
        interactive = False
    else:
        interactive = True

    while True:
        try:
            if interactive:
                message = input("> ").strip()

                try:
                    # This lets users hit the up arrow key for past messages
                    readline.add_history(message)
                except:
                    # If the user doesn't have readline (may be the case on windows), that's fine
                    pass

        except KeyboardInterrupt:
            # Exit gracefully
            break

        if message.startswith("%") and interactive:
            handle_magic_command(cosmo, message)
            continue

        # Many users do this
        if message.strip() == "cosmo --local":
            print("Please press CTRL-C then run `cosmo --local`.")
            continue

        # Track if we've ran a code block.
        # We'll use this to determine if we should render a new code block,
        # In the event we get code -> output -> code again
        ran_code_block = False
        render_cursor = True

        try:
            for chunk in cosmo.chat(message, display=False, stream=True):
                if cosmo.debug_mode:
                    print("Chunk in `terminal_interface`:", chunk)

                # Message
                if "message" in chunk:
                    if active_block is None:
                        active_block = MessageBlock()
                    if active_block.type != "message":
                        active_block.end()
                        active_block = MessageBlock()
                    active_block.message += chunk["message"]
                    render_cursor = True

                # Code
                if "code" in chunk or "language" in chunk:
                    if active_block is None:
                        active_block = CodeBlock()
                    if active_block.type != "code" or ran_code_block:
                        # If the last block wasn't a code block,
                        # or it was, but we already ran it:
                        active_block.end()
                        active_block = CodeBlock()
                    ran_code_block = False
                    render_cursor = True

                if "language" in chunk:
                    active_block.language = chunk["language"]
                if "code" in chunk:
                    active_block.code += chunk["code"]
                if "active_line" in chunk:
                    active_block.active_line = chunk["active_line"]

                # Execution notice
                if "executing" in chunk:
                    if not cosmo.auto_run:
                        # OI is about to execute code. The user wants to approve this

                        # End the active block so you can run input() below it
                        active_block.end()

                        should_scan_code = False

                        if not cosmo.safe_mode == "off":
                            if cosmo.safe_mode == "auto":
                                should_scan_code = True
                            elif cosmo.safe_mode == "ask":
                                response = input(
                                    "  Would you like to scan this code? (y/n)\n\n  "
                                )
                                print("")  # <- Aesthetic choice

                                if response.strip().lower() == "y":
                                    should_scan_code = True

                        if should_scan_code:
                            # Get code language and actual code from the chunk
                            # We need to give these to semgrep when we start our scan
                            language = chunk["executing"]["language"]
                            code = chunk["executing"]["code"]

                            scan_code(code, language, cosmo)

                        response = input(
                            "  Would you like to run this code? (y/n)\n\n  "
                        )
                        print("")  # <- Aesthetic choice

                        if response.strip().lower() == "y":
                            # Create a new, identical block where the code will actually be run
                            # Conveniently, the chunk includes everything we need to do this:
                            active_block = CodeBlock()
                            active_block.margin_top = False  # <- Aesthetic choice
                            active_block.language = chunk["executing"]["language"]
                            active_block.code = chunk["executing"]["code"]
                        else:
                            # User declined to run code.
                            cosmo.messages.append(
                                {
                                    "role": "user",
                                    "message": "I have declined to run this code.",
                                }
                            )
                            break

                # Output
                if "output" in chunk:
                    ran_code_block = True
                    render_cursor = False
                    active_block.output += "\n" + chunk["output"]
                    active_block.output = (
                        active_block.output.strip()
                    )  # <- Aesthetic choice

                    # Truncate output
                    active_block.output = truncate_output(
                        active_block.output, cosmo.max_output
                    )

                if active_block:
                    active_block.refresh(cursor=render_cursor)

                yield chunk

            # (Sometimes -- like if they CTRL-C quickly -- active_block is still None here)
            if active_block:
                active_block.end()
                active_block = None

            if not interactive:
                # Don't loop
                break

        except KeyboardInterrupt:
            # Exit gracefully
            if active_block:
                active_block.end()
                active_block = None

            if interactive:
                # (this cancels LLM, returns to the interactive "> " input)
                continue
            else:
                break



import sys

from .core.core import Cosmo

# This is done so when users `import cosmo`,
# they get an instance of cosmo:

sys.modules["cosmo"] = Cosmo()

# **This is a controversial thing to do,**
# because perhaps modules ought to behave like modules.

# But I think it saves a step, removes friction, and looks good.




import json


def convert_to_openai_messages(messages, function_calling=True):
    new_messages = []

    for message in messages:
        new_message = {"role": message["role"], "content": ""}

        if "message" in message:
            new_message["content"] = message["message"]

        if "code" in message:
            if function_calling:
                new_message["function_call"] = {
                    "name": "execute",
                    "arguments": json.dumps(
                        {"language": message["language"], "code": message["code"]}
                    ),
                    # parsed_arguments isn't actually an OpenAI thing, it's an OI thing.
                    # but it's soo useful! we use it to render messages to text_llms
                    "parsed_arguments": {
                        "language": message["language"],
                        "code": message["code"],
                    },
                }
            else:
                new_message[
                    "content"
                ] += f"""\n\n```{message["language"]}\n{message["code"]}\n```"""
                new_message["content"] = new_message["content"].strip()

        new_messages.append(new_message)

        if "output" in message:
            if function_calling:
                new_messages.append(
                    {
                        "role": "function",
                        "name": "execute",
                        "content": message["output"],
                    }
                )
            else:
                new_messages.append(
                    {
                        "role": "assistant",
                        "content": "The code above was executed on the user's machine. Output: "
                        + message["output"],
                    }
                )

    return new_messages



import json
import re


def merge_deltas(original, delta):
    """
    Pushes the delta into the original and returns that.

    Great for reconstructing OpenAI streaming responses -> complete message objects.
    """
    for key, value in delta.items():
        if isinstance(value, dict):
            if key not in original:
                original[key] = value
            else:
                merge_deltas(original[key], value)
        else:
            if key in original:
                original[key] += value
            else:
                original[key] = value
    return original



import os

import litellm

from .convert_to_coding_llm import convert_to_coding_llm
from .setup_openai_coding_llm import setup_openai_coding_llm
from .setup_text_llm import setup_text_llm


def setup_llm(cosmo):
    """
    Takes a Cosmo (which includes a ton of LLM settings),
    returns a Coding LLM (a generator that streams deltas with `message` and `code`).
    """

    if not cosmo.local and (
        cosmo.model in litellm.open_ai_chat_completion_models
        or cosmo.model.startswith("azure/")
    ):
        # Function-calling LLM
        coding_llm = setup_openai_coding_llm(cosmo)
    else:
        # Non-function-calling LLM
        text_llm = setup_text_llm(cosmo)
        coding_llm = convert_to_coding_llm(text_llm, debug_mode=cosmo.debug_mode)

    return coding_llm



import getpass
import os
import time

import litellm

from ..utils.display_markdown_message import display_markdown_message


def validate_llm_settings(cosmo):
    """
    Interactivley prompt the user for required LLM settings
    """

    # This runs in a while loop so `continue` lets us start from the top
    # after changing settings (like switching to/from local)
    while True:
        if cosmo.local:
            # We have already displayed a message.
            # (This strange behavior makes me think validate_llm_settings needs to be rethought / refactored)
            break

        else:
            # Ensure API keys are set as environment variables

            # OpenAI
            if cosmo.model in litellm.open_ai_chat_completion_models:
                if not os.environ.get("OPENAI_API_KEY") and not cosmo.api_key:
                    display_welcome_message_once()

                    display_markdown_message(
                        """---
                    > OpenAI API key not found

                    To use `GPT-4` (highly recommended) please provide an OpenAI API key.

                    To use another language model, Please refer to our documentation at [docs.cosmoai.net](https://docs.cosmoai.net/language-model-setup/).
                    
                    ---
                    """
                    )

                    response = getpass.getpass("OpenAI API key: ")
                    print(f"OpenAI API key: {response[:4]}...{response[-4:]}")

                    display_markdown_message(
                        """

                    **Tip:** To save this key for later, run `export OPENAI_API_KEY=your_api_key` on Mac/Linux or `setx OPENAI_API_KEY your_api_key` on Windows.
                    
                    ---"""
                    )

                    cosmo.api_key = response
                    time.sleep(2)
                    break

            # This is a model we don't have checks for yet.
            break

    # If we're here, we passed all the checks.

    # Auto-run is for fast, light useage -- no messages.
    # If local, we've already displayed a message.
    if not cosmo.auto_run and not cosmo.local:
        display_markdown_message(f"> Model set to `{cosmo.model}`")
    return


def display_welcome_message_once():
    """
    Displays a welcome message only on its first call.

    (Uses an internal attribute `_displayed` to track its state.)
    """
    if not hasattr(display_welcome_message_once, "_displayed"):
        display_markdown_message(
            """
        ●

        Welcome to **Cosmo AI**.

        We're glad to have you here. Cosmo AI is a powerful tool designed to help you with you just anything you want!

        If you need any assistance, please refer to our documentation at [docs.cosmoai.net](https://docs.cosmoai.net/). 

        Happy Coding!
        """
        )
        time.sleep(1.5)

        display_welcome_message_once._displayed = True



import os
import tempfile
import webbrowser

from ..base_code_cosmo import BaseCodeCosmo


class HTML(BaseCodeCosmo):
    file_extension = "html"
    proper_name = "HTML"

    def __init__(self):
        super().__init__()

    def run(self, code):
        # Create a temporary HTML file with the content
        with tempfile.NamedTemporaryFile(delete=False, suffix=".html") as f:
            f.write(code.encode())

        # Open the HTML file with the default web browser
        webbrowser.open("file://" + os.path.realpath(f.name))

        yield {
            "output": f"Saved to {os.path.realpath(f.name)} and opened with the user's default web browser."
        }



class BaseCodeCosmo:
    """
    .run is a generator that yields a dict with attributes: active_line, output
    """

    def __init__(self):
        pass

    def run(self, code):
        pass

    def terminate(self):
        pass



import argparse
import os
import platform
import subprocess

import pkg_resources

from ..terminal_interface.conversation_navigator import conversation_navigator
from ..utils.display_markdown_message import display_markdown_message
from ..utils.get_config import get_config_path

arguments = [
    {
        "name": "system_message",
        "nickname": "s",
        "help_text": "prompt / custom instructions for the language model",
        "type": str,
    },
    {
        "name": "local",
        "nickname": "l",
        "help_text": "experimentally run the language model locally (via LM Studio)",
        "type": bool,
    },
    {
        "name": "auto_run",
        "nickname": "y",
        "help_text": "automatically run generated code",
        "type": bool,
    },
    {
        "name": "debug_mode",
        "nickname": "d",
        "help_text": "run in debug mode",
        "type": bool,
    },
    {
        "name": "model",
        "nickname": "m",
        "help_text": "language model to use",
        "type": str,
    },
    {
        "name": "temperature",
        "nickname": "t",
        "help_text": "optional temperature setting for the language model",
        "type": float,
    },
    {
        "name": "context_window",
        "nickname": "c",
        "help_text": "optional context window size for the language model",
        "type": int,
    },
    {
        "name": "max_tokens",
        "nickname": "x",
        "help_text": "optional maximum number of tokens for the language model",
        "type": int,
    },
    {
        "name": "max_budget",
        "nickname": "b",
        "help_text": "optionally set the max budget (in USD) for your llm calls",
        "type": float,
    },
    {
        "name": "api_base",
        "nickname": "ab",
        "help_text": "optionally set the API base URL for your llm calls (this will override environment variables)",
        "type": str,
    },
    {
        "name": "api_key",
        "nickname": "ak",
        "help_text": "optionally set the API key for your llm calls (this will override environment variables)",
        "type": str,
    },
    {
        "name": "safe_mode",
        "nickname": "safe",
        "help_text": "optionally enable safety mechanisms like code scanning; valid options are off, ask, and auto",
        "type": str,
        "choices": ["off", "ask", "auto"],
        "default": "off",
    },
    {
        "name": "config_file",
        "nickname": "cf",
        "help_text": "optionally set a custom config file to use",
        "type": str,
    },
]


def cli(cosmo):
    parser = argparse.ArgumentParser(description="Cosmo")

    # Add arguments
    for arg in arguments:
        if arg["type"] == bool:
            parser.add_argument(
                f'-{arg["nickname"]}',
                f'--{arg["name"]}',
                dest=arg["name"],
                help=arg["help_text"],
                action="store_true",
                default=None,
            )
        else:
            choices = arg["choices"] if "choices" in arg else None
            default = arg["default"] if "default" in arg else None

            parser.add_argument(
                f'-{arg["nickname"]}',
                f'--{arg["name"]}',
                dest=arg["name"],
                help=arg["help_text"],
                type=arg["type"],
                choices=choices,
                default=default,
            )

    # Add special arguments
    parser.add_argument(
        "--config",
        dest="config",
        action="store_true",
        help="open config.yaml file in text editor",
    )
    parser.add_argument(
        "--conversations",
        dest="conversations",
        action="store_true",
        help="list conversations to resume",
    )
    parser.add_argument(
        "-f",
        "--fast",
        dest="fast",
        action="store_true",
        help="run `cosmo --model gpt-3.5-turbo`",
    )
    parser.add_argument(
        "--version",
        dest="version",
        action="store_true",
        help="get Cosmo's version number",
    )

    args = parser.parse_args()

    # This should be pushed into an open_config.py util
    # If --config is used, open the config.yaml file in the Cosmo folder of the user's config dir
    if args.config:
        if args.config_file:
            config_file = get_config_path(args.config_file)
        else:
            config_file = get_config_path()

        print(f"Opening `{config_file}`...")

        # Use the default system editor to open the file
        if platform.system() == "Windows":
            os.startfile(
                config_file
            )  # This will open the file with the default application, e.g., Notepad
        else:
            try:
                # Try using xdg-open on non-Windows platforms
                subprocess.call(["xdg-open", config_file])
            except FileNotFoundError:
                # Fallback to using 'open' on macOS if 'xdg-open' is not available
                subprocess.call(["open", config_file])
        return

    if args.local:
        # Default local (LM studio) attributes
        cosmo.system_message = "You are an AI."
        cosmo.model = (
            "openai/" + cosmo.model
        )  # This tells LiteLLM it's an OpenAI compatible server
        cosmo.api_base = "http://localhost:1234/v1"
        cosmo.max_tokens = 4000
        cosmo.context_window = 128000
        cosmo.api_key = "0"

        display_markdown_message(
            """
> Cosmo's local mode is powered by **`LM Studio`**.


You will need to run **LM Studio** in the background.

1. Download **LM Studio** from [https://lmstudio.ai/](https://lmstudio.ai/) then start it.
2. Select a language model then click **Download**.
3. Click the **<->** button on the left (below the chat button).
4. Select your model at the top, then click **Start Server**.


Once the server is running, you can begin your conversation below.

> **Warning:** This feature is highly experimental.
> Don't expect `gpt-3.5` / `gpt-4` level quality, speed, or reliability yet!

"""
        )

    # Set attributes on cosmo
    for attr_name, attr_value in vars(args).items():
        # Ignore things that aren't possible attributes on cosmo
        if attr_value is not None and hasattr(cosmo, attr_name):
            # If the user has provided a config file, load it and extend cosmo's configuration
            if attr_name == "config_file":
                user_config = get_config_path(attr_value)
                cosmo.config_file = user_config
                cosmo.extend_config(config_path=user_config)
            else:
                setattr(cosmo, attr_name, attr_value)

    # if safe_mode and auto_run are enabled, safe_mode disables auto_run
    if cosmo.auto_run and (
        cosmo.safe_mode == "ask" or cosmo.safe_mode == "auto"
    ):
        setattr(cosmo, "auto_run", False)

    # If --conversations is used, run conversation_navigator
    if args.conversations:
        conversation_navigator(cosmo)
        return

    if args.version:
        version = pkg_resources.get_distribution("cosmo").version
        print(f"Cosmo {version}")
        return

    if args.fast:
        cosmo.model = "gpt-3.5-turbo"

    cosmo.chat()
